// rationalize.cpp 2016-02-03 	multidimensional chain fractions

#include <math.h>
#include "rationalize.h"

static double vdbl_len(int N, double *a)				// |a|
{
	double res = 0;
	for (int i = N; i>0; i--)
	{
		double d1 = (*a++);
		res += d1*d1;
	}
	return sqrt(res);
}
static double vdbl_scpod(int N, double *a, double *b)	// (a,b)
{
	double res = 0;
	for (int i = N; i>0; i--)
		res += (*a++)*(*b++);
	return res;
}
static void vdbl_add(int N, double *a, double *b, double R)	// a += R*b
{
	for (int i = N; i>0; i--)
		(*a++) += R*(*b++);
}
static double vdbl_minAbs(int N, double *a, int &i)		// minimal abs.value of non-zero elemtnts = a[i]
{
	i = -1;
	double min = 1e99;
	for (int j = 0; j < N; j++) {
		double aj = fabs(a[j]);
		if (aj == 0) continue;
		if (aj < min) { i = j; min = aj; }
	}
	return min;
}
static void   vdbl_zero(int N, double *a, double eps)	// less<eps -> 0
{
	for (int i = N; i > 0; i--, a++)
		if (fabs(*a) < eps) *a = 0;
}
static double deviationI(double R)						// deviation R from integer
{
	return fabs(R - floor(R + 0.5));
}
static double	dSquare(double x) { return x*x; };
static double vdbl_devi(int n, double *a, double co)	// deviation a*co from integer
{
	double su = 0;
	for (int i = 0; i<n; i++)
	{
		su += dSquare(deviationI(a[i] * co));
	}
	return sqrt(su);
}
//----------------------------------------------------------------------------
double rationalize(int n, double *v_, double tol)		// calculate R: v[i]*R - (almost) integer with deviation tol. Return R
{
	int y, yMin;
	double len, len1, fMin, fMi1, *aMin;
	len = vdbl_len(n, v_);
	if (len == 0) return 0;
	len = 1 / len;
	double *v = new double[n];
	for (y = 0; y<n; y++) v[y] = len*v_[y];
	// now len(v) = 1
	const double eps = 1e-13;
	vdbl_zero(n, v, eps);					// all positive values, less then eps would be 0
	
	double *a = new double[n*n];			// square matrix a of size n*n
	double *t = a;
	for (y = n*n; y > 0; --y) *t++ = 0;
	for (y = 0; y < n; y++) a[n*y + y] = 1;	// a := identical matrix

	double *f = new double[n];				// vector of size n
	for (y = 0; y < n; y++) f[y] = v[y];	// f := v

	for (int iStep = 0; iStep < 10; iStep++) {
		vdbl_minAbs(n, f, yMin);
		fMin = f[yMin];
		fMi1 = 1 / fMin;
		aMin = a + n*yMin;
		for (y = 0; y < n; y++) {
			if (y == yMin) continue;
			double k = floor(f[y] * fMi1 + 0.5);
			if (k == 0) continue;
			vdbl_add(n, aMin, a + n*y, k);	// row(yMin) += k*row(y)
			f[y] -= k*fMin;
		}
		vdbl_zero(n, aMin, eps);
		// if scalar product <0, negate vector
		if (vdbl_scpod(n, aMin, v)< 0) {
			for (int y = 0; y < n; y++) aMin[y] *= -1;
			f[yMin] *= -1;
		}
		len1 = vdbl_len(n, aMin);
		if (vdbl_devi(n, v, len1)<tol*len1) break;
	}
	delete[]v;
	delete[]f;
	delete[]a;
	return len1*len;
}

